/* -*- tab-width: 4 -*-
 *
 * Electric(tm) VLSI Design System
 *
 * File: EGraphics.java
 *
 * Copyright (c) 2003 Sun Microsystems and Static Free Software
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 */
package com.sun.electric.database.geometry;

import com.sun.electric.database.text.Pref;
import com.sun.electric.database.text.TextUtils;
import com.sun.electric.technology.Layer;
import com.sun.electric.technology.Technology;

import java.awt.Color;
import java.util.HashMap;

/**
 * Class to define the appearance of a piece of geometry.
 */
public class EGraphics
{
	/** the Layer associated with this graphics. */			private Layer layer;
	/** display: true to use patterns; false for solid */	private boolean displayPatterned;
	/** display: true to outline patterns */				private boolean displayOutlined;
	/** printer: true to use patterns; false for solid */	private boolean printPatterned;
	/** printer: true to outline patterns */				private boolean printOutlined;
	/** transparent layer to use (0 for none) */			private int transparentLayer;
	/** color to use */										private int red, green, blue;
	/** opacity (0 to 1) of color */						private double opacity;
	/** whether to draw color in foregound */				private int foreground;
	/** stipple pattern to draw */							private int [] pattern;

	private static HashMap usePatternDisplayMap = new HashMap();
	private static HashMap outlinePatternPrinterMap = new HashMap();
	private static HashMap usePatternPrinterMap = new HashMap();
	private static HashMap outlinePatternDisplayMap = new HashMap();
	private static HashMap transparentLayerMap = new HashMap();
	private static HashMap colorMap = new HashMap();
	private static HashMap patternMap = new HashMap();

	/**
	 * There are 3 ways to encode color in an integer.
	 * If the lowest bit (FULLRGBBIT) is set, this is a full RGB color in the high 3 bytes.
	 * If the next lowest bit (OPAQUEBIT) is set, this is an "old C Electric" opaque color
	 * (such as WHITE, BLACK, etc., listed below).
	 * If neither of these bits is set, this is a transparent layer
	 * (LAYERT1, LAYERT2, etc., listed below).
	 */
	/** Describes the full RGB escape bit. */				public final static int FULLRGBBIT = 01;
	/** Describes opaque color escape bit. */				public final static int OPAQUEBIT =  02;

	// the opaque colors (all have the OPAQUEBIT set in them)
	/** Describes the color white. */						public final static int WHITE =    0002;
	/** Describes the color black. */						public final static int BLACK =    0006;
	/** Describes the color red. */							public final static int RED =      0012;
	/** Describes the color blue. */						public final static int BLUE =     0016;
	/** Describes the color green. */						public final static int GREEN =    0022;
	/** Describes the color cyan. */						public final static int CYAN =     0026;
	/** Describes the color magenta. */						public final static int MAGENTA =  0032;
	/** Describes the color yellow. */						public final static int YELLOW =   0036;
	/** Describes the cell and port names. */				public final static int CELLTXT =  0042;
	/** Describes the cell outline. */						public final static int CELLOUT =  0046;
	/** Describes the window border color. */				public final static int WINBOR =   0052;
	/** Describes the highlighted window border color. */	public final static int HWINBOR =  0056;
	/** Describes the menu border color. */					public final static int MENBOR =   0062;
	/** Describes the highlighted menu border color. */		public final static int HMENBOR =  0066;
	/** Describes the menu text color. */					public final static int MENTXT =   0072;
	/** Describes the menu glyph color. */					public final static int MENGLY =   0076;
	/** Describes the cursor color. */						public final static int CURSOR =   0102;
	/** Describes the color gray. */						public final static int GRAY =     0106;
	/** Describes the color orange. */						public final static int ORANGE =   0112;
	/** Describes the color purple. */						public final static int PURPLE =   0116;
	/** Describes the color brown. */						public final static int BROWN =    0122;
	/** Describes the color light gray. */					public final static int LGRAY =    0126;
	/** Describes the color dark gray. */					public final static int DGRAY =    0132;
	/** Describes the color light red. */					public final static int LRED =     0136;
	/** Describes the color dark red. */					public final static int DRED =     0142;
	/** Describes the color light green. */					public final static int LGREEN =   0146;
	/** Describes the color dark green. */					public final static int DGREEN =   0152;
	/** Describes the color light blue. */					public final static int LBLUE =    0156;
	/** Describes the color dark blue. */					public final static int DBLUE =    0162;

	// the transparent layers
	/** Describes transparent layer 1. */					public final static int LAYERT1 =      04;
	/** Describes transparent layer 2. */					public final static int LAYERT2 =     010;
	/** Describes transparent layer 3. */					public final static int LAYERT3 =     020;
	/** Describes transparent layer 4. */					public final static int LAYERT4 =     040;
	/** Describes transparent layer 5. */					public final static int LAYERT5 =    0100;
	/** Describes transparent layer 6. */					public final static int LAYERT6 =    0200;
	/** Describes transparent layer 7. */					public final static int LAYERT7 =    0400;
	/** Describes transparent layer 8. */					public final static int LAYERT8 =   01000;
	/** Describes transparent layer 9. */					public final static int LAYERT9 =   02000;
	/** Describes transparent layer 10. */					public final static int LAYERT10 =  04000;
	/** Describes transparent layer 11. */					public final static int LAYERT11 = 010000;
	/** Describes transparent layer 12. */					public final static int LAYERT12 = 020000;

	// Constants used in technologies and in creating an EGraphics
	/** defines the 1st transparent layer. */				public static final int TRANSPARENT_1 = 1;
	/** defines the 2nd transparent layer. */				public static final int TRANSPARENT_2 = 2;
	/** defines the 3rd transparent layer. */				public static final int TRANSPARENT_3 = 3;
	/** defines the 4th transparent layer. */				public static final int TRANSPARENT_4 = 4;
	/** defines the 5th transparent layer. */				public static final int TRANSPARENT_5 = 5;
	/** defines the 6th transparent layer. */				public static final int TRANSPARENT_6 = 6;
	/** defines the 7th transparent layer. */				public static final int TRANSPARENT_7 = 7;
	/** defines the 8th transparent layer. */				public static final int TRANSPARENT_8 = 8;
	/** defines the 9th transparent layer. */				public static final int TRANSPARENT_9 = 9;
	/** defines the 10th transparent layer. */				public static final int TRANSPARENT_10 = 10;
	/** defines the 11th transparent layer. */				public static final int TRANSPARENT_11 = 11;
	/** defines the 12th transparent layer. */				public static final int TRANSPARENT_12 = 12;

	// drawing styles
	/** Draw as a solid fill. */							public final static int SOLID      = 0;
	/** Draw as a stipple pattern. */						public final static int PATTERNED  = 1;
	/** Draw as a stipple pattern with an outline. */		public final static int OUTLINEPAT = 2;

	/**
	 * Method to create a graphics object.
	 * @param displayMethod the way to show this EGraphics on a display (SOLID, PATTERNED, or OUTLINEPAT).
	 * @param printMethod the way to show this EGraphics on paper (SOLID, PATTERNED, or OUTLINEPAT).
	 * @param transparentLayer the transparent layer number (0 for none).
	 * @param red the red component of this EGraphics.
	 * @param green the green component of this EGraphics.
	 * @param blue the blue component of this EGraphics.
	 * @param opacity the opacity of this EGraphics (1 for opaque, 0 for transparent).
	 * @param foreground the foreground factor of this EGraphics (1 for to be in foreground).
	 * @param pattern the 16x16 stipple pattern of this EGraphics (16 integers).
	 * This pattern is tessellated across the polygon.
	 */
	public EGraphics(int displayMethod, int printMethod, int transparentLayer,
		int red, int green, int blue, double opacity, int foreground, int[] pattern)
	{
		this.layer = null;
		this.displayPatterned = (displayMethod != SOLID);
		this.displayOutlined = (displayMethod == OUTLINEPAT);
		this.printPatterned = (printMethod != SOLID);
		this.printOutlined = (printMethod == OUTLINEPAT);
		this.transparentLayer = transparentLayer;
		this.red = red;
		this.green = green;
		this.blue = blue;
		this.opacity = opacity;
		this.foreground = foreground;
		this.pattern = pattern;
		if (pattern.length != 16)
		{
			System.out.println("Graphics bad: has " + pattern.length + " pattern entries instead of 16");
		}
		if (transparentLayer < 0 || transparentLayer >= TRANSPARENT_12)
		{
			System.out.println("Graphics transparent color bad: " + transparentLayer);
		}
		if (red < 0 || red > 255 || green < 0 || green > 255 || blue < 0 || blue > 255)
		{
			System.out.println("Graphics color bad: (" + red + "," + green + "," + blue + ")");
		}
	}

	/**
	 * Method tells which Layer is associated with this EGraphics.
	 * @return the Layer that is associated with this EGraphics.
	 * Returns null if no Layer is associated.
	 */
	public Layer getLayer() { return layer; }

	/**
	 * Method to set which Layer is associated with this EGraphics.
	 * Since this is called only during initialization, it also examines preferences
	 * and uses them to override the graphics information.
	 * @param layer the Layer to associate with this EGraphics.
	 */
	public void setLayer(Layer layer)
	{
		this.layer = layer;
		Technology tech = layer.getTechnology();
		if (tech == null) return;

		Pref usePatternDisplayPref = Pref.makeBooleanPref("UsePatternDisplayFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), displayPatterned);
		displayPatterned = usePatternDisplayPref.getBoolean();
		usePatternDisplayMap.put(layer, usePatternDisplayPref);

		Pref outlinePatternDisplayPref = Pref.makeBooleanPref("OutlinePatternDisplayFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), displayOutlined);
		displayOutlined = outlinePatternDisplayPref.getBoolean();
		outlinePatternDisplayMap.put(layer, outlinePatternDisplayPref);

		Pref usePatternPrinterPref = Pref.makeBooleanPref("UsePatternPrinterFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), printPatterned);
		printPatterned = usePatternPrinterPref.getBoolean();
		usePatternPrinterMap.put(layer, usePatternPrinterPref);

		Pref outlinePatternPrinterPref = Pref.makeBooleanPref("OutlinePatternPrinterFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), printOutlined);
		printOutlined = outlinePatternPrinterPref.getBoolean();
		outlinePatternPrinterMap.put(layer, outlinePatternPrinterPref);

		Pref transparentLayerPref = Pref.makeIntPref("TransparentLayerFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), transparentLayer);
		transparentLayer = transparentLayerPref.getInt();
		transparentLayerMap.put(layer, transparentLayerPref);

		Pref colorPref = Pref.makeIntPref("ColorFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), (red<<16) | (green << 8) | blue);
		int color = colorPref.getInt();
		red = (color >> 16) & 0xFF;
		green = (color >> 8) & 0xFF;
		blue = color & 0xFF;
		colorMap.put(layer, colorPref);

		String pat = makePatString(pattern);
		Pref patternPref = Pref.makeStringPref("PatternFor" + layer.getName() + "In" + tech.getTechName(),
			Technology.getTechnologyPreferences(), pat);
		pat = patternPref.getString();
		parsePatString(pat, pattern);
		patternMap.put(layer, patternPref);
	}

	private String makePatString(int [] pattern)
	{
		StringBuffer sb = new StringBuffer();
		for(int i=0; i<16; i++)
		{
			if (i > 0) sb.append("/");
			sb.append(Integer.toString(pattern[i]));
		}
		return sb.toString();
	}

	private void parsePatString(String patString, int [] pattern)
	{
		int pos = 0;
		for(int i=0; i<16; i++)
		{
			pattern[i] = TextUtils.atoi(patString.substring(pos));
			pos = patString.indexOf('/', pos) + 1;
		}
	}

	/**
	 * Method describes how this EGraphics appears on a display.
	 * This EGraphics can be drawn as a solid fill or as a pattern.
	 * @return true to draw this EGraphics patterned on a display.
	 * False to draw this EGraphics as a solid fill on a display.
	 */
	public boolean isPatternedOnDisplay() { return displayPatterned; }

	/**
	 * Method to set how this EGraphics appears on a display.
	 * This EGraphics can be drawn as a solid fill or as a pattern.
	 * @param p true to draw this EGraphics patterned on a display.
	 * False to draw this EGraphics as a solid fill on a display.
	 */
	public void setPatternedOnDisplay(boolean p)
	{
		displayPatterned = p;

		if (layer != null)
		{
			Pref pref = (Pref)usePatternDisplayMap.get(layer);
			if (pref != null) pref.setBoolean(p);
		}
	}

	/**
	 * Method to tell whether this pattern has an outline around it on the display.
	 * When the EGraphics is drawn as a pattern, the outline can be defined more clearly by drawing a line around the edge.
	 * @return true to dan outline around this pattern (on the display).
	 */
	public boolean isOutlinedOnDisplay() { return displayOutlined; }

	/**
	 * Method to set whether this pattern has an outline around it on the display.
	 * When the EGraphics is drawn as a pattern, the outline can be defined more clearly by drawing a line around the edge.
	 * @param o true to draw this pattern with an outline around it.
	 */
	public void setOutlinedOnDisplay(boolean o)
	{
		displayOutlined = o;

		if (layer != null)
		{
			Pref pref = (Pref)outlinePatternDisplayMap.get(layer);
			if (pref != null) pref.setBoolean(o);
		}
	}

	/**
	 * Method describes how this EGraphics appears on a printer.
	 * This EGraphics can be drawn as a solid fill or as a pattern.
	 * @return true to draw this EGraphics patterned on a printer.
	 * False to draw this EGraphics as a solid fill on a printer.
	 */
	public boolean isPatternedOnPrinter() { return printPatterned; }

	/**
	 * Method to set how this EGraphics appears on a printer.
	 * This EGraphics can be drawn as a solid fill or as a pattern.
	 * @param p true to draw this EGraphics patterned on a printer.
	 * False to draw this EGraphics as a solid fill on a printer.
	 */
	public void setPatternedOnPrinter(boolean p)
	{
		this.printPatterned = p;

		if (layer != null)
		{
			Pref pref = (Pref)usePatternPrinterMap.get(layer);
			if (pref != null) pref.setBoolean(p);
		}
	}

	/**
	 * Method to tell whether this pattern has an outline around it.
	 * When the EGraphics is drawn as a pattern, the outline can be defined more clearly by drawing a line around the edge.
	 * @return true to dan outline around this pattern (on the printer).
	 */
	public boolean isOutlinedOnPrinter() { return printOutlined; }

	/**
	 * Method to set whether this pattern has an outline around it on a printer.
	 * When the EGraphics is drawn as a pattern, the outline can be defined more clearly by drawing a line around the edge.
	 * @param o true to draw this pattern with an outline around it on a printer.
	 */
	public void setOutlinedOnPrinter(boolean o)
	{
		printOutlined = o;

		if (layer != null)
		{
			Pref pref = (Pref)outlinePatternPrinterMap.get(layer);
			if (pref != null) pref.setBoolean(o);
		}
	}

	/**
	 * Method to return the transparent layer number associated with this EGraphics.
	 * @return the transparent layer number associated with this EGraphics.
	 * A value of zero means that this EGraphics is not drawn transparently.
	 * Instead, use the "getColor()" method to get its solid color.
	 */
	public int getTransparentLayer() { return transparentLayer; }

	/**
	 * Method to set the transparent layer number associated with this EGraphics.
	 * @param transparentLayer the transparent layer number associated with this EGraphics.
	 * A value of zero means that this EGraphics is not drawn transparently.
	 * Then, use the "setColor()" method to set its solid color.
	 */
	public void setTransparentLayer(int transparentLayer)
	{
		this.transparentLayer = transparentLayer;

		if (layer != null)
		{
			Pref pref = (Pref)transparentLayerMap.get(layer);
			if (pref != null) pref.setInt(transparentLayer);
		}
	}

	/**
	 * Method to get the stipple pattern of this EGraphics.
	 * The stipple pattern is a 16 x 16 pattern that is stored in 16 integers.
	 * @return the stipple pattern of this EGraphics.
	 */
	public int [] getPattern() { return pattern; }

	/**
	 * Method to set the stipple pattern of this EGraphics.
	 * The stipple pattern is a 16 x 16 pattern that is stored in 16 integers.
	 * @param pattern the stipple pattern of this EGraphics.
	 */
	public void setPattern(int [] pattern)
	{
		this.pattern = pattern;

		if (layer != null)
		{
			Pref pref = (Pref)patternMap.get(layer);
			if (pref != null) pref.setString(makePatString(pattern));
		}
	}

	/**
	 * Method to get the opacity of this EGraphics.
	 * Opacity runs from 0 (transparent) to 1 (opaque).
	 * @return the opacity of this EGraphics.
	 */
	public double getOpacity() { return opacity; }

	/**
	 * Method to set the opacity of this EGraphics.
	 * Opacity runs from 0 (transparent) to 1 (opaque).
	 * @param opacity the opacity of this EGraphics.
	 */
	public void setOpacity(double opacity) { this.opacity = opacity; }

	/**
	 * Method to get whether this EGraphics should be drawn in the foreground.
	 * The foreground is the main "mix" of layers, such as metals and polysilicons.
	 * The background is typically used by implant and well layers.
	 * @return the whether this EGraphics should be drawn in the foreground.
	 */
	public boolean getForeground() { return foreground != 0; }

	/**
	 * Method to return the color associated with this EGraphics.
	 * @return the color associated with this EGraphics.
	 */
	public Color getColor()
	{
		int alpha = (int)(opacity * 255.0);
		Color color = new Color(red, green, blue, alpha);
		return color;
	}

	/**
	 * Method to set the color associated with this EGraphics.
	 * @param color the color to set.
	 */
	public void setColor(Color color)
	{
		transparentLayer = 0;
		red = color.getRed();
		green = color.getGreen();
		blue = color.getBlue();
		if (layer != null)
		{
			Pref pref = (Pref)colorMap.get(layer);
			if (pref != null) pref.setInt((red << 16) | (green << 8) | blue);
		}
	}

	/**
	 * Method to set the color index associated with this EGraphics.
	 * Color indices are more general than colors, because they can handle
	 * the old C-Electric color values as well as full color values.
	 * Artwork nodes and arcs represent individualized color by using color indices.
	 * @param color the color index to set.
	 */
	public void setColorIndex(int color)
	{
		if ((color&OPAQUEBIT) != 0)
		{
			// an opaque color
			transparentLayer = 0;
			switch (color)
			{
				case WHITE:   setColor(new Color(255, 255, 255));   break;
				case BLACK:   setColor(new Color(  0,   0,   0));   break;
				case RED:     setColor(new Color(255,   0,   0));   break;
				case BLUE:    setColor(new Color(  0,   0, 255));   break;
				case GREEN:   setColor(new Color(  0, 255,   0));   break;
				case CYAN:    setColor(new Color(  0, 255, 255));   break;
				case MAGENTA: setColor(new Color(255,   0, 255));   break;
				case YELLOW:  setColor(new Color(255, 255,   0));   break;
				case CELLTXT: setColor(new Color(  0,   0,   0));   break;
				case CELLOUT: setColor(new Color(  0,   0,   0));   break;
				case WINBOR:  setColor(new Color(  0,   0,   0));   break;
				case HWINBOR: setColor(new Color(  0, 255,   0));   break;
				case MENBOR:  setColor(new Color(  0,   0,   0));   break;
				case HMENBOR: setColor(new Color(255, 255, 255));   break;
				case MENTXT:  setColor(new Color(  0,   0,   0));   break;
				case MENGLY:  setColor(new Color(  0,   0,   0));   break;
				case CURSOR:  setColor(new Color(  0,   0,   0));   break;
				case GRAY:    setColor(new Color(180, 180, 180));   break;
				case ORANGE:  setColor(new Color(255, 190,   6));   break;
				case PURPLE:  setColor(new Color(186,   0, 255));   break;
				case BROWN:   setColor(new Color(139,  99,  46));   break;
				case LGRAY:   setColor(new Color(230, 230, 230));   break;
				case DGRAY:   setColor(new Color(100, 100, 100));   break;
				case LRED:    setColor(new Color(255, 150, 150));   break;
				case DRED:    setColor(new Color(159,  80,  80));   break;
				case LGREEN:  setColor(new Color(175, 255, 175));   break;
				case DGREEN:  setColor(new Color( 89, 159,  85));   break;
				case LBLUE:   setColor(new Color(150, 150, 255));   break;
				case DBLUE:   setColor(new Color(  2,  15, 159));   break;
			}
			return;
		}
		if ((color&FULLRGBBIT) != 0)
		{
			// a full RGB color (opaque)
			transparentLayer = 0;
			setColor(new Color((color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF));
			return;
		}

		// a transparent color
		if ((color&LAYERT1) != 0) transparentLayer = TRANSPARENT_1; else
		if ((color&LAYERT2) != 0) transparentLayer = TRANSPARENT_2; else
		if ((color&LAYERT3) != 0) transparentLayer = TRANSPARENT_3; else
		if ((color&LAYERT4) != 0) transparentLayer = TRANSPARENT_4; else
		if ((color&LAYERT5) != 0) transparentLayer = TRANSPARENT_5; else
		if ((color&LAYERT6) != 0) transparentLayer = TRANSPARENT_6; else
		if ((color&LAYERT7) != 0) transparentLayer = TRANSPARENT_7; else
		if ((color&LAYERT8) != 0) transparentLayer = TRANSPARENT_8; else
		if ((color&LAYERT9) != 0) transparentLayer = TRANSPARENT_9; else
		if ((color&LAYERT10) != 0) transparentLayer = TRANSPARENT_10; else
		if ((color&LAYERT11) != 0) transparentLayer = TRANSPARENT_11; else
		if ((color&LAYERT12) != 0) transparentLayer = TRANSPARENT_12;
	}

	/**
	 * Method to convert a Color to a color index.
	 * Color indices are more general than colors, because they can handle
	 * the old C-Electric color values as well as full color values.
	 * Artwork nodes and arcs represent individualized color by using color indices.
	 * @param color a Color object
	 * @return the color index that describes that color.
	 */
	public static int makeIndex(Color color)
	{
		int red =   color.getRed();
		int green = color.getGreen();
		int blue =  color.getBlue();
		int index = (red << 24) | (green << 16) | (blue << 8) | FULLRGBBIT;
		return index;
	}

	/**
	 * Method to convert a transparent layer to a color index.
	 * Color indices are more general than colors, because they can handle
	 * the old C-Electric color values as well as full color values.
	 * One of the extensions they also have is the ability to describe the transarent layers.
	 * @param transparentLayer the transparent layer number.
	 * @return the color index that describes that transparent layer.
	 */
	public static int makeIndex(int transparentLayer)
	{
		switch (transparentLayer)
		{
			case TRANSPARENT_1: return LAYERT1;
			case TRANSPARENT_2: return LAYERT2;
			case TRANSPARENT_3: return LAYERT3;
			case TRANSPARENT_4: return LAYERT4;
			case TRANSPARENT_5: return LAYERT5;
			case TRANSPARENT_6: return LAYERT6;
			case TRANSPARENT_7: return LAYERT7;
			case TRANSPARENT_8: return LAYERT8;
			case TRANSPARENT_9: return LAYERT9;
			case TRANSPARENT_10: return LAYERT10;
			case TRANSPARENT_11: return LAYERT11;
			case TRANSPARENT_12: return LAYERT12;
		}
		return 0;
	}

	/**
	 * Method to find the index of a color, given its name.
	 * @param name the name of the color.
	 * @return the index of the color.
	 */
	public static int findColorIndex(String name)
	{
		if (name.equals("white"))          return WHITE;
		if (name.equals("black"))          return BLACK;
		if (name.equals("red"))            return RED;
		if (name.equals("blue"))           return BLUE;
		if (name.equals("green"))          return GREEN;
		if (name.equals("cyan"))           return CYAN;
		if (name.equals("magenta"))        return MAGENTA;
		if (name.equals("yellow"))         return YELLOW;
		if (name.equals("gray"))           return GRAY;
		if (name.equals("orange"))         return ORANGE;
		if (name.equals("purple"))         return PURPLE;
		if (name.equals("brown"))          return BROWN;
		if (name.equals("light-gray"))     return LGRAY;
		if (name.equals("dark-gray"))      return DGRAY;
		if (name.equals("light-red"))      return LRED;
		if (name.equals("dark-red"))       return DRED;
		if (name.equals("light-green"))    return LGREEN;
		if (name.equals("dark-green"))     return DGREEN;
		if (name.equals("light-blue"))     return LBLUE;
		if (name.equals("dark-blue"))      return DBLUE;
		if (name.equals("transparent-1"))  return LAYERT1;
		if (name.equals("transparent-2"))  return LAYERT2;
		if (name.equals("transparent-3"))  return LAYERT3;
		if (name.equals("transparent-4"))  return LAYERT4;
		if (name.equals("transparent-5"))  return LAYERT5;
		if (name.equals("transparent-6"))  return LAYERT6;
		if (name.equals("transparent-7"))  return LAYERT7;
		if (name.equals("transparent-8"))  return LAYERT8;
		if (name.equals("transparent-9"))  return LAYERT9;
		if (name.equals("transparent-10")) return LAYERT10;
		if (name.equals("transparent-11")) return LAYERT11;
		if (name.equals("transparent-12")) return LAYERT12;
		return 0;
	}

	/**
	 * Method to tell the name of the color with a given index.
	 * Color indices are more general than colors, because they can handle
	 * the old C-Electric color values as well as full color values.
	 * @param color the color number.
	 * @return the name of that color.
	 */
	public static String getColorIndexName(int color)
	{
		if ((color&FULLRGBBIT) != 0)
		{
			int red =   (color >> 24) & 0xFF;
			int green = (color >> 16) & 0xFF;
			int blue =  (color >> 8) & 0xFF;
			return "Color (" + red + "," + green + "," + blue + ")";
		}
		switch (color)
		{
			case WHITE:    return "white";
			case BLACK:    return "black";
			case RED:      return "red";
			case BLUE:     return "blue";
			case GREEN:    return "green";
			case CYAN:     return "cyan";
			case MAGENTA:  return "magenta";
			case YELLOW:   return "yellow";
			case GRAY:     return "gray";
			case ORANGE:   return "orange";
			case PURPLE:   return "purple";
			case BROWN:    return "brown";
			case LGRAY:    return "light-gray";
			case DGRAY:    return "dark-gray";
			case LRED:     return "light-red";
			case DRED:     return "dark-red";
			case LGREEN:   return "light-green";
			case DGREEN:   return "dark-green";
			case LBLUE:    return "light-blue";
			case DBLUE:    return "dark-blue";
			case LAYERT1:  return "transparent-1";
			case LAYERT2:  return "transparent-2";
			case LAYERT3:  return "transparent-3";
			case LAYERT4:  return "transparent-4";
			case LAYERT5:  return "transparent-5";
			case LAYERT6:  return "transparent-6";
			case LAYERT7:  return "transparent-7";
			case LAYERT8:  return "transparent-8";
			case LAYERT9:  return "transparent-9";
			case LAYERT10: return "transparent-10";
			case LAYERT11: return "transparent-11";
			case LAYERT12: return "transparent-12";
		}
		return "Color "+color;
	}

	/**
	 * Method to return the array of colors.
	 * @return an array of the possible colors.
	 */
	public static int [] getColorIndices()
	{
		return new int [] {WHITE, BLACK, RED, BLUE, GREEN, CYAN, MAGENTA, YELLOW,
			GRAY, ORANGE, PURPLE, BROWN, LGRAY, DGRAY, LRED, DRED, LGREEN, DGREEN, LBLUE, DBLUE,
			LAYERT1, LAYERT2, LAYERT3, LAYERT4, LAYERT5, LAYERT6, LAYERT7, LAYERT8, LAYERT9,
			LAYERT10, LAYERT11, LAYERT12};
	}

	/**
	 * Method to return the array of transparent colors.
	 * @return an array of the possible transparent colors.
	 */
	public static int [] getTransparentColorIndices()
	{
		return new int [] {LAYERT1, LAYERT2, LAYERT3, LAYERT4, LAYERT5, LAYERT6, LAYERT7, LAYERT8, LAYERT9,
			LAYERT10, LAYERT11, LAYERT12};
	}
}
